searchData={"content_type":"text/markdown","items":[{"doc":"Read model projections for Commanded using Ecto.","ref":"Commanded.Projections.Ecto.html","title":"Commanded.Projections.Ecto","type":"behaviour"},{"doc":"defmodule Projector do\n      use Commanded.Projections.Ecto,\n        application: MyApp.Application,\n        name: \"my-projection\",\n        repo: MyApp.Repo,\n        schema_prefix: \"my-prefix\",\n        timeout: :infinity\n\n      project %Event{}, _metadata, fn multi ->\n        Ecto.Multi.insert(multi, :my_projection, %MyProjection{...})\n      end\n\n      project %AnotherEvent{}, fn multi ->\n        Ecto.Multi.insert(multi, :my_projection, %MyProjection{...})\n      end\n    end","ref":"Commanded.Projections.Ecto.html#module-example-usage","title":"Example usage - Commanded.Projections.Ecto","type":"behaviour"},{"doc":"- [Getting started](getting-started.html)\n- [Usage](usage.html)","ref":"Commanded.Projections.Ecto.html#module-guides","title":"Guides - Commanded.Projections.Ecto","type":"behaviour"},{"doc":"The optional `after_update/3` callback function defined in a projector is\ncalled after each projected event.\n\nThe function receives the event, its metadata, and all changes from the\n`Ecto.Multi` struct that were executed within the database transaction.\n\nYou could use this function to notify subscribers that the read model has been\nupdated, such as by publishing changes via Phoenix PubSub channels.","ref":"Commanded.Projections.Ecto.html#c:after_update/3","title":"Commanded.Projections.Ecto.after_update/3","type":"callback"},{"doc":"defmodule MyApp.ExampleProjector do\n      use Commanded.Projections.Ecto,\n        application: MyApp.Application,\n        repo: MyApp.Projections.Repo,\n        name: \"MyApp.ExampleProjector\"\n\n      project %AnEvent{name: name}, fn multi ->\n        Ecto.Multi.insert(multi, :example_projection, %ExampleProjection{name: name})\n      end\n\n      @impl Commanded.Projections.Ecto\n      def after_update(event, metadata, changes) do\n        # Use the event, metadata, or `Ecto.Multi` changes and return `:ok`\n        :ok\n      end\n    end","ref":"Commanded.Projections.Ecto.html#c:after_update/3-example","title":"Example - Commanded.Projections.Ecto.after_update/3","type":"callback"},{"doc":"Project a domain event into a read model by appending one or more operations\nto the `Ecto.Multi` struct passed to the projection function you define\n\nThe operations will be executed in a database transaction including an\nidempotency check to guarantee an event cannot be projected more than once.","ref":"Commanded.Projections.Ecto.html#project/2","title":"Commanded.Projections.Ecto.project/2","type":"macro"},{"doc":"project %AnEvent{}, fn multi ->\n      Ecto.Multi.insert(multi, :my_projection, %MyProjection{...})\n    end","ref":"Commanded.Projections.Ecto.html#project/2-example","title":"Example - Commanded.Projections.Ecto.project/2","type":"macro"},{"doc":"Project a domain event and its metadata map into a read model by appending one\nor more operations to the `Ecto.Multi` struct passed to the projection\nfunction you define.\n\nThe operations will be executed in a database transaction including an\nidempotency check to guarantee an event cannot be projected more than once.","ref":"Commanded.Projections.Ecto.html#project/3","title":"Commanded.Projections.Ecto.project/3","type":"macro"},{"doc":"project %AnEvent{}, metadata, fn multi ->\n      Ecto.Multi.insert(multi, :my_projection, %MyProjection{...})\n    end","ref":"Commanded.Projections.Ecto.html#project/3-example","title":"Example - Commanded.Projections.Ecto.project/3","type":"macro"},{"doc":"The optional `schema_prefix/1` callback function defined in a projector is\nused to set the schema of the `projection_versions` table used by the\nprojector for idempotency checks.\n\nIt is passed the event and its metadata and must return the schema name, as a\nstring, or `nil`.","ref":"Commanded.Projections.Ecto.html#c:schema_prefix/1","title":"Commanded.Projections.Ecto.schema_prefix/1","type":"callback"},{"doc":"The optional `schema_prefix/2` callback function defined in a projector is\nused to set the schema of the `projection_versions` table used by the\nprojector for idempotency checks.\n\nIt is passed the event and its metadata, and must return the schema name, as a\nstring, or `nil`","ref":"Commanded.Projections.Ecto.html#c:schema_prefix/2","title":"Commanded.Projections.Ecto.schema_prefix/2","type":"callback"},{"doc":"# Changelog","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"#","ref":"changelog.html#1-4-0","title":"1.4.0 - Changelog","type":"extras"},{"doc":"- Use an UPSERT SQL query to insert or update the projection version ([#52](https://github.com/commanded/commanded-ecto-projections/pull/52)).","ref":"changelog.html#enhancements","title":"Enhancements - Changelog","type":"extras"},{"doc":"#","ref":"changelog.html#1-3-0","title":"1.3.0 - Changelog","type":"extras"},{"doc":"- Fix Elixir 1.14 compilation warnings ([#45](https://github.com/commanded/commanded-ecto-projections/pull/45)).\n\n---","ref":"changelog.html#enhancements","title":"Enhancements - Changelog","type":"extras"},{"doc":"#","ref":"changelog.html#1-2-1","title":"1.2.1 - Changelog","type":"extras"},{"doc":"- Allow exceptions to be rescued by Commanded's event handler ([#37](https://github.com/commanded/commanded-ecto-projections/pull/37)).","ref":"changelog.html#enhancements","title":"Enhancements - Changelog","type":"extras"},{"doc":"#","ref":"changelog.html#1-2-0","title":"1.2.0 - Changelog","type":"extras"},{"doc":"- Support runtime projector names ([#32](https://github.com/commanded/commanded-ecto-projections/pull/32)).\n- Support `schema_prefix/2` function ([#33](https://github.com/commanded/commanded-ecto-projections/pull/33)).\n\n---","ref":"changelog.html#enhancements","title":"Enhancements - Changelog","type":"extras"},{"doc":"#","ref":"changelog.html#1-1-0","title":"1.1.0 - Changelog","type":"extras"},{"doc":"- Dynamic schema prefix ([#28](https://github.com/commanded/commanded-ecto-projections/pull/28)).\n- Support Commanded v1.1.0.\n\n---","ref":"changelog.html#enhancements","title":"Enhancements - Changelog","type":"extras"},{"doc":"#","ref":"changelog.html#1-0-0","title":"1.0.0 - Changelog","type":"extras"},{"doc":"- Support multiple Commanded apps ([#25](https://github.com/commanded/commanded-ecto-projections/pull/25)).\n- Add `.formatter.exs` to Hex package ([#19](https://github.com/commanded/commanded-ecto-projections/pull/19)).\n- Add microseconds to timestamp fields in `projection_versions` ([#22](https://github.com/commanded/commanded-ecto-projections/pull/22)).\n\n---","ref":"changelog.html#enhancements","title":"Enhancements - Changelog","type":"extras"},{"doc":"#","ref":"changelog.html#0-8-0","title":"0.8.0 - Changelog","type":"extras"},{"doc":"- Upgrade to Ecto v3 ([#17](https://github.com/commanded/commanded-ecto-projections/pull/17)).\n- Use lambda instead of unhygienic var in projection macros ([#13](https://github.com/commanded/commanded-ecto-projections/pull/13)).\n\n  Previously _magic_ `multi`:\n\n  ```elixir\n  project %AnEvent{name: name}, _metadata do\n    Ecto.Multi.insert(multi, :example_projection, %ExampleProjection{name: name})\n  end\n  ```\n\n  Now `multi` is provided as a argument to the project function:\n\n  ```elixir\n  project %AnEvent{name: name}, _metadata, fn multi ->\n    Ecto.Multi.insert(multi, :example_projection, %ExampleProjection{name: name})\n  end\n  ```\n\n  The previous `do` block approach is still supported, but has been deprecated. It will be removed in the next release.\n\n---","ref":"changelog.html#enhancements","title":"Enhancements - Changelog","type":"extras"},{"doc":"#","ref":"changelog.html#0-7-1","title":"0.7.1 - Changelog","type":"extras"},{"doc":"- Ensure errors encountered while building the `Ecto.Multi` data structure within a `project` function are caught and passed to the `error/3` callback.","ref":"changelog.html#bug-fixes","title":"Bug fixes - Changelog","type":"extras"},{"doc":"#","ref":"changelog.html#0-7-0","title":"0.7.0 - Changelog","type":"extras"},{"doc":"- Support Commanded's event handler `error/3` callback ([#12](https://github.com/commanded/commanded-ecto-projections/pull/12)).\n\n---","ref":"changelog.html#enhancements","title":"Enhancements - Changelog","type":"extras"},{"doc":"#","ref":"changelog.html#0-6-0","title":"0.6.0 - Changelog","type":"extras"},{"doc":"- Pass through any additional projector configuration options to Commanded event handler.\n  Allows new Commanded features to be used without updating this library (e.g. specify `consistency` option).\n\n---","ref":"changelog.html#enhancements","title":"Enhancements - Changelog","type":"extras"},{"doc":"#","ref":"changelog.html#0-5-0","title":"0.5.0 - Changelog","type":"extras"},{"doc":"- Allow an Ecto schema prefix to be defined in config or per handler ([#4](https://github.com/commanded/commanded-ecto-projections/pull/4)).\n\n---","ref":"changelog.html#enhancements","title":"Enhancements - Changelog","type":"extras"},{"doc":"#","ref":"changelog.html#0-4-0","title":"0.4.0 - Changelog","type":"extras"},{"doc":"- Add `repo` option to `Commanded.Projections.Ecto` macro ([#1](https://github.com/commanded/commanded-ecto-projections/pull/1)).\n- Optional `after_update/3` callback function in projectors.","ref":"changelog.html#enhancements","title":"Enhancements - Changelog","type":"extras"},{"doc":"# Getting started\n\nYou should already have [Ecto](https://github.com/elixir-ecto/ecto) installed and configured before proceeding. Please follow Ecto's [Getting Started](https://hexdocs.pm/ecto/getting-started.html) guide to get going first.\n\n1. Add `:commanded_ecto_projections` to your list of dependencies in `mix.exs`:\n\n    ```elixir\n    def deps do\n      [\n        {:commanded_ecto_projections, \"~> 1.4\"}\n      ]\n    end\n    ```\n\n2. Generate an Ecto migration in your app:\n\n    ```console\n    $ mix ecto.gen.migration create_projection_versions\n    ```\n\n3. Modify the generated migration, in `priv/repo/migrations`, to create the `projection_versions` table:\n\n    ```elixir\n    defmodule CreateProjectionVersions do\n      use Ecto.Migration\n\n      def change do\n        create table(:projection_versions, primary_key: false) do\n          add(:projection_name, :text, primary_key: true)\n          add(:last_seen_event_number, :bigint)\n\n          timestamps(type: :naive_datetime_usec)\n        end\n      end\n    end\n    ```\n\n5. Run the Ecto migration:\n\n    ```console\n    $ mix ecto.migrate\n    ```\n\n6. Define your first read model projector:\n\n    ```elixir\n    defmodule MyApp.ExampleProjector do\n      use Commanded.Projections.Ecto,\n        application: MyApp.Application,\n        repo: MyApp.Projections.Repo,\n        name: \"example_projection\"\n    end\n    ```\n\nRefer to the Usage guide for more detail on how to configure and use a read model projector.","ref":"getting-started.html","title":"Getting started","type":"extras"},{"doc":"# Usage","ref":"usage.html","title":"Usage","type":"extras"},{"doc":"Use `Ecto.Schema` to define one or more read models:\n\n```elixir\ndefmodule ExampleProjection do\n  use Ecto.Schema\n\n  schema \"example_projections\" do\n    field(:name, :string)\n  end\nend\n```","ref":"usage.html#creating-a-read-model","title":"Creating a read model - Usage","type":"extras"},{"doc":"For each read model you will need to define a module that uses the `Commanded.Projections.Ecto` module and projects the appropriate domain events with the `project` macro.\n\nYou must specify the following options when defining or starting an Ecto projector:\n\n- `:application` - (module or atom) the Commanded application (e.g. `MyApp.Application`).\n- `:name` - (string) a unique name used to identify the event store subscription used by the projector.\n- `:repo` - (module) an Ecto repo (e.g. `MyApp.Projections.Repo`).\n\nOnce a projector has been deployed you _should not_ change its name. Doing so will cause a new event store subscription to be created and replay all existing events.\n\n**Note:** A read model projector is just a specialised Commanded event handler `GenServer` process.\n\n#","ref":"usage.html#creating-a-projector","title":"Creating a projector - Usage","type":"extras"},{"doc":"```elixir\ndefmodule MyApp.ExampleProjector do\n  use Commanded.Projections.Ecto,\n    application: MyApp.Application,\n    repo: MyApp.Projections.Repo,\n    name: \"example_projection\"\n\n  project %AnEvent{name: name}, _metadata, fn multi ->\n    Ecto.Multi.insert(multi, :example_projection, %ExampleProjection{name: name})\n  end\n\n  project %AnotherEvent{name: name}, fn multi ->\n    Ecto.Multi.insert(multi, :example_projection, %ExampleProjection{name: name})\n  end\nend\n```\n\n##","ref":"usage.html#example","title":"Example - Usage","type":"extras"},{"doc":"The `:application` and `:name` options can be provided at runtime, but `:repo` must be specified at compile-time.\n\n```elixir\ndefmodule MyApp.ExampleProjector do\n  use Commanded.Projections.Ecto,\n    repo: MyApp.Projections.Repo\nend\n```\n\nStarted with:\n\n```elixir\n{:ok, pid} = ExampleProjector.start_link(application: MyApp.Application, name: \"example_projection\")\n```\n\nOr supervised:\n\n```elixir\nSupervisor.start_link([\n  {ExampleProjector, application: MyApp.Application, name: \"example_projection\"}\n], strategy: :one_for_one)\n```\n\nRuntime configuration allows the same projector to be run more than once, with each instance using a separate application or name:\n\n```elixir\nSupervisor.start_link([\n  {ExampleProjector, application: App1, name: \"App1.Projector\"},\n  {ExampleProjector, application: App2, name: \"App2.Projector\"}\n], strategy: :one_for_one)\n```\n\n#","ref":"usage.html#runtime-configuration","title":"Runtime configuration - Usage","type":"extras"},{"doc":"The `project/3` macro expects the domain event, metadata, and a single-arity function that takes and returns an `Ecto.Multi` data structure for grouping multiple Repo operations. These will all be executed within a single transaction. You can use `Ecto.Multi` to insert, update, and delete data.\n\n##","ref":"usage.html#using-the-project-macro","title":"Using the `project` macro - Usage","type":"extras"},{"doc":"Project an event and its metadata into a read model with `project/3`:\n\n```elixir\nproject %AnEvent{name: name}, metadata, fn multi ->\n  projection = %ExampleProjection{name: name, metadata: metadata}\n\n  Ecto.Multi.insert(multi, :example_projection, projection)\nend\n```\n\nUse `project/2` if you do not need to use the event metadata:\n\n```elixir\nproject %AnotherEvent{name: name}, fn multi ->\n  Ecto.Multi.insert(multi, :example_projection, %ExampleProjection{name: name})\nend\n```\n\nIf you want to skip a projection event, you can return the `multi` transaction without further modifying it:\n\n```elixir\nproject %ItemUpdated{uuid: uuid} = event, _metadata, fn multi ->\n  case Repo.get(ItemProjection, uuid) do\n    nil -> multi\n    item -> Ecto.Multi.update(multi, :item, update_changeset(event, item))\n  end\nend\n```","ref":"usage.html#examples","title":"Examples - Usage","type":"extras"},{"doc":"Your projector module must be included in your application supervision tree:\n\n```elixir\ndefmodule MyApp.Projections.Supervisor do\n  use Supervisor\n\n  def start_link(init_arg) do\n    Supervisor.start_link(__MODULE__, init_arg, name: __MODULE__)\n  end\n\n  @impl true\n  def init(_init_arg) do\n    children = [\n      MyApp.ExampleProjector\n    ]\n\n    Supervisor.init(children, strategy: :one_for_one)\n  end\nend\n```\n\n**Warning:** You should implement an [error handling](#error-handling) strategy in your projector module when supervising to prevent problematic events from causing cascading errors due too many restarts.","ref":"usage.html#supervision","title":"Supervision - Usage","type":"extras"},{"doc":"### `error/3` callback\n\nThe `Commanded.Projections.Ecto` macro defines a Commanded event handler which means you can take advantage of the [`error/3` callback function](https://hexdocs.pm/commanded/Commanded.Event.Handler.html#module-error-3-callback) to handle any errors returned from a `project` function. The error function is passed the error returned by the event handler (e.g. `{:error, error}`), the event causing the error, and a context map containing state passed between retries. Use the context map to track any transient state you need to access between retried failures, such as the number of failed attempts.\n\nYou can return one of the following responses depending upon the error severity:\n\n- `{:retry, context}` - retry the failed event, provide a context map containing any state passed to subsequent failures. This could be used to count the number of failures, stopping after too many.\n\n- `{:retry, delay, context}` - retry the failed event, after sleeping for the requested delay (in milliseconds). Context is a map as described in `{:retry, context}` above.\n\n- `:skip` - skip the failed event by acknowledging receipt.\n\n- `{:stop, reason}` - stop the projector with the given reason.\n\n##","ref":"usage.html#error-handling","title":"Error handling - Usage","type":"extras"},{"doc":"Here's an example projector module where an error tagged tuple is explicitly returned from a `project` function, but you can also handle exceptions caused by faulty `Ecto.Multi` database operations in a similar manner since the errors are caught and returned as tagged tuples (e.g. `{:error, %Ecto.ConstraintError{}}`).\n\n```elixir\ndefmodule MyApp.ExampleProjector do\n  use Commanded.Projections.Ecto,\n    application: MyApp.Application,\n    repo: MyApp.Projections.Repo,\n    name: \"MyApp.ExampleProjector\"\n\n  require Logger\n\n  alias Commanded.Event.FailureContext\n\n  project %AnEvent{}, fn _multi ->\n    {:error, :failed}\n  end\n\n  def error({:error, :failed}, %AnEvent{}, %FailureContext{}) do\n    :skip\n  end\n\n  def error({:error, %Ecto.ConstraintError{} = error}, _event, _failure_context) do\n    Logger.error(\"Failed due to constraint error: \" <> inspect(error))\n\n    :skip\n  end\n\n  def error({:error, _error}, _event, _failure_context) do\n    :skip\n  end\nend\n```\n\n### `after_update/3` callback\n\nYou can define an `after_update/3` callback function in a projector to be called after each projected event. The function receives the event, its metadata, and all changes from the `Ecto.Multi` struct that were executed within the database transaction.\n\n```elixir\ndefmodule MyApp.ExampleProjector do\n  use Commanded.Projections.Ecto,\n    application: MyApp.Application,\n    repo: MyApp.Projections.Repo,\n    name: \"MyApp.ExampleProjector\"\n\n  project %AnEvent{name: name}, fn multi ->\n    Ecto.Multi.insert(multi, :example_projection, %ExampleProjection{name: name})\n  end\n\n  @impl Commanded.Projections.Ecto\n  def after_update(event, metadata, changes) do\n    # Use the event, metadata, or `Ecto.Multi` changes and return `:ok`\n    :ok\n  end\nend\n```\n\nYou could use this function to notify subscribers that the read model has been updated (e.g. pub/sub to Phoenix channels).","ref":"usage.html#error-handling-example","title":"Error handling example - Usage","type":"extras"},{"doc":"When using a prefix for your Ecto schemas you might also want to change the prefix for the `ProjectionVersion` schema. There are a number of options to do this:\n\n1. Define a global static prefix via environment config:\n\n    ```elixir\n    # config/config.exs\n    config :commanded_ecto_projections, schema_prefix: \"example_schema_prefix\"\n    ```\n\n2. Provide a static `schema_prefix` as a projector option:\n\n    ```elixir\n    defmodule MyApp.ExampleProjector do\n      use Commanded.Projections.Ecto,\n        application: MyApp.Application,\n        repo: MyApp.Projections.Repo,\n        name: \"example_projection\",\n        schema_prefix: \"example_schema_prefix\"\n    end\n    ```\n\n3. Provide a one-arity function as a `schema_prefix` projector option:\n\n    ```elixir\n    defmodule MyApp.ExampleProjector do\n      use Commanded.Projections.Ecto,\n        application: MyApp.Application,\n        repo: MyApp.Projections.Repo,\n        name: \"example_projection\",\n        schema_prefix: fn event -> \"example_schema_prefix\" end\n    end\n    ```\n\n    The function will receive the event as the single argument allowing you to use the same or a different schema for each event.\n\n4. Provide a two-arity function as a `schema_prefix` projector option:\n\n    ```elixir\n    defmodule MyApp.ExampleProjector do\n      use Commanded.Projections.Ecto,\n        application: MyApp.Application,\n        repo: MyApp.Projections.Repo,\n        name: \"example_projection\",\n        schema_prefix: fn event, metadata -> \"example_schema_prefix\" end\n    end\n    ```\n\n    The function will receive the event and its associated metadata as the two arguments allowing you to use the same or a different schema for each event. The metadata will also include the enriched fields such as the application, event handler name, and optional handler state.\n\n5. Define a `schema_prefix/1` callback function:\n\n    ```elixir\n    defmodule MyApp.ExampleProjector do\n      use Commanded.Projections.Ecto,\n        application: MyApp.Application,\n        name: \"example_projection\"\n\n      @impl Commanded.Projections.Ecto\n      def schema_prefix(event), do: \"example_schema_prefix\"\n    end\n    ```\n\n    The function will receive the event as the single argument allowing you to use the same or a different schema for each event.\n\n    An example usage could be for tenant specific projections where each tenant's data is projected and stored in a separate database schema:\n\n    ```elixir\n    @impl Commanded.Projections.Ecto\n    def schema_prefix(%_{tenant: tenant}), do: tenant\n    ```\n\n6. Define a `schema_prefix/2` callback function:\n\n    ```elixir\n    defmodule MyApp.ExampleProjector do\n      use Commanded.Projections.Ecto,\n        application: MyApp.Application,\n        name: \"example_projection\"\n\n      @impl Commanded.Projections.Ecto\n      def schema_prefix(event, metadata), do: \"example_schema_prefix\"\n    end\n    ```\n\n    The function will receive the event and its associated metadata as the two arguments allowing you to use the same or a different schema for each event. The metadata will also include the enriched fields such as the application, event handler name, and optional handler state.\n\n    An example usage could be for tenant specific projections where each tenant's data is projected and stored in a separate database schema:\n\n    ```elixir\n    @impl Commanded.Projections.Ecto\n    def schema_prefix(%_{tenant: tenant}, _metadata), do: tenant\n    ```\n\n#","ref":"usage.html#schema-prefix","title":"Schema prefix - Usage","type":"extras"},{"doc":"1. Generate an Ecto migration in your app:\n\n    ```shell\n    mix ecto.gen.migration create_schema_projection_versions\n    ```\n\n2. Modify the generated migration, in `priv/repo/migrations`, to create the schema and a `projection_versions` table for the schema:\n\n    ```elixir\n    defmodule CreateSchemaProjectionVersions do\n      use Ecto.Migration\n\n      def up do\n        execute(\"CREATE SCHEMA example_schema_prefix\")\n\n        create table(:projection_versions, primary_key: false, prefix: \"example_schema_prefix\") do\n          add(:projection_name, :text, primary_key: true)\n          add(:last_seen_event_number, :bigint)\n\n          timestamps(type: :naive_datetime_usec)\n        end\n      end\n\n      def down do\n        drop(table(:projection_versions, prefix: \"example_schema_prefix\"))\n\n        execute(\"DROP SCHEMA example_schema_prefix CASCADE\")\n      end\n    end\n    ```\n\n    Note you will need to do this for each schema prefix you use.","ref":"usage.html#migrations-with-a-schema-prefix","title":"Migrations with a schema prefix - Usage","type":"extras"},{"doc":"The `projection_versions` table is used to ensure that events are only projected once.\n\nTo rebuild a projection you will need to:\n\n1. Delete the row containing the last seen event for the projection name:\n\n    ```SQL\n    DELETE FROM projection_versions\n    WHERE projection_name = 'example_projection';\n    ```\n\n2. Truncate the tables that are being populated by the projection, and restart their identity:\n\n    ```SQL\n    TRUNCATE TABLE\n      example_projections,\n      other_projections\n    RESTART IDENTITY;\n    ```\n\nYou will also need to reset the event store subscription for the commanded event handler. This is specific to whichever event store you are using.","ref":"usage.html#rebuilding-a-projection","title":"Rebuilding a projection - Usage","type":"extras"}]}